; Q68 PTR_GEN mouse driver. Base on code by Wolfgang Lenerz

	section mouse

	include 'dev8_keys_con'
	include 'dev8_keys_err'
	include 'dev8_keys_sys'
	include 'dev8_keys_iod'
	include 'dev8_keys_chn'
	include 'dev8_keys_qdos_sms'
	include 'dev8_keys_qdos_ioa'
	include 'dev8_keys_qdos_io'
	include 'dev8_keys_qlv'
	include 'dev8_keys_q68'

; Mouse driver working data
ms_lext   equ	$00	; Must be at offset 0
ms_aext   equ	$04
ms_con	  equ	$08	; CON driver definition block
ms_state  equ	$0C
ms_btn	  equ	$0E
ms_xmov   equ	$10
ms_ymov   equ	$12
ms_scrl   equ	$14
ms.size   equ	$18

md_start
	moveq	#sms.info,d0
	trap	#1
	movea.l a0,a6			; sysVars in a6

	moveq	#ioa.open,d0
	moveq	#-1,d1
	moveq	#0,d3
	lea	con_name(pc),a0
	trap	#2			; Open CON channel
	tst.l	d0
	bne.s	md_exit

	moveq	#iop.pinf,d0
	trap	#3
	tst.l	d0
	bne.s	err_pinf

; Get channel block
	move.l	a0,d2
	move.l	sys_chtb(a6),a1
	lsl.w	#2,d2
	move.l	(a1,d2.w),a1		; CDB
	move.l	chn_drvr(a1),a1 	; CON driver
	lea	-pt_liod(a1),a4 	; A4 = CON driver definition block

	moveq	#ioa.clos,d0
	trap	#2			; Close channel again

	moveq	#sms.achp,d0
	moveq	#ms.size,d1
	moveq	#0,d2
	moveq	#0,d3
	trap	#1
	tst.l	d0
	bne.s	md_exit
	move.l	a0,a3			; This is now the mouse data block

	move.l	a4,ms_con(a3)		; Remember CON linkage block
	move.w	#2,ms_state(a3) 	; State always starts at 2

	lea	pt68_ext(pc),a1 	; Set interrupt routine address
	move.l	a1,ms_aext(a3)
	lea	ms_lext(a3),a0		; Link in interrupt routine
	moveq	#sms.lexi,d0
	trap	#1
md_exit
	rts

err_pinf
	moveq	#ioa.clos,d0
	trap	#2			; Close channel again

	lea	pinf_txt,a1
	move.l	a0,-(a7)
	suba.l	a0,a0
	movea.w ut.wtext,a2
	jsr	(a2)
	move.l	(sp)+,a0
	moveq	#err.fdnf,d0
	rts

; External interrupt
pt68_ext
	move.b	mouse_status,d3
	btst	#m.rcv,d3		; any mouse activity?
	beq.s	mse_done		; no, must have been another int ->

	move.l	ms_con(a3),a5		; CON linkage block
	clr.l	d1
	move.b	mouse_code,d1		; get mouse code
	ext.w	d1			; mouse code as a word
	lea	ms_state(a3),a2 	; mouse status area
	move.w	(a2),d0 		; current mouse status
	move.w	d1,(a2,d0.w)		; store this value
	addq.w	#2,d0			; next place to store value
	btst	#m.type,d3		; wheel mouse?
	bne.s	whl_mse 		; yes

	cmp.b	#8,d0			; no, got 3rd byte?
	bne.s	mse_add 		; no, not done yet
	clr.w	ms_scrl(a3)		; there is no scroll
	bra.s	mse_pckt		; treat packet now

whl_mse cmp.w	#10,d0			; got the entire packet?
	beq.s	mse_pckt		; yes ->

mse_add move.w	d0,(a2) 		; no
	st	mouse_unlock		; release mouse receive status
mse_done
	rts

mse_pckt
; here a (3 or 4 byte) mouse packet is complete.
	move.w	#2,(a2) 		; prepare next packet

	move.l	ms_xmov(a3),d2		; X|Y mouse move
	beq.s	mse_scr 		; no move, check scroll
	move.w	d2,d1			; y move
	swap	d2			; x move
	neg.w	d1			; apparently vertical mvt is inverted
; note d1 and d2 are switched here, compared to std code
	move.w	d2,d0			; supposed count
	bpl.s	mse_ycount		; ... it's positive
	neg.w	d0
mse_ycount
	move.w	d1,d3			; other supposed count
	bge.s	mse_addinc		; ... it's positive
	neg.w	d3
mse_addinc
	add.w	d3,d0			; total count
	beq.s	mse_setinc		; ... none
	add.w	pt_xicnt(a5),d0 	; total so far
	lsr.w	#1,d0			; reduced
	add.b	pt_wake(a5),d0		; wake up speed
	lsr.w	#1,d0
	addq.w	#1,d0			; make sure that there is a move
mse_setinc
	move.w	d0,pt_xicnt(a5) 	; update count
	add.w	d2,pt_xinc(a5)
	add.w	d1,pt_yinc(a5)		; and distance moved

; now check mouse scrollwheel
mse_scr
	move.w	ms_btn(a3),d2		; mouse button
;;;;bra.s mse_btn
;;;	   move.w  ms_scrl(a3),d1	   ; mouse wheelmove
;;;	   andi.w  #15,d1		   ;
;;;	   move.w  d1,d3		   ; mouse wheel distance moved
;;;	   beq.s   mse_btn		   ; no mouse wheel activity
;;;	   move.l  sys_ckyq(a6),d0	   ; current kbd queue
;;;	   beq.s   mse_btn		   ; none
;;;	   move.l  d0,a2
;;;	   move.w  #$D1D9,d1		   ; ALT+dwn/ALT+up
;;;	   move.l  sys_klnk(a6),d0
;;;	   beq.s   mse_wct
;;;	   move.l  d0,a3		   ; keyboard linkage
;;;	   btst    #kb..ctrl,kb_stat(a3)   ; CTRL pressed?
;;;	   beq.s   mse_wct		   ; ...no
;;;	   move.w  #$C1C9,d1		   ; ALT+right/ALT+left
;;;mse_wct
;;;	   btst    #3,d3		   ; negative?
;;;	   beq.s   mse_wup		   ; no->
;;;	   ror.w   #8,d1
;;;	   or.b    #$f0,d3
;;;	   neg.b   d3

;;;mse_wup subq.w  #1,d3		   ; prepare for dbf
;;;mse_wlp bsr.s   ioq_pbyt		   ; byte into kbd queue
;;;	   bne.s   mse_btn
;;;	   dbf	   d3,mse_wlp

; now check mouse buttons
mse_btn 				; d2.w = button clicked
	andi.b	#7,d2			; only keep lower 3 bits
	cmp.b	#2,d2			; more than right/left click ?
	bgt.s	cntr			; yes, this is button 3, do centre button
cnt	move.b	d2,pt_bpoll(a5) 	; this is new button or old button still depressed
	beq.s	ms_exit 		; ... none
	clr.b	pt_lstuf(a5)		; it isn't a stuff
ms_exit
	st	mouse_unlock		; release mouse receive status
	rts

cntr
; this is a modified copy of pt_button3
; on entry
; a3 = ptr to my mouse status area (below)
; a5 = kbd linkage
; a6 = sysvars
	move.l	sys_ckyq(a6),a2 	; stuff in here
	tas	pt_lstuf(a5)		; was last press a stuff? (this one is)
	blt.s	ms_exit 		; yes, don't do another
	move.b	pt_stuf1(a5),d1 	; get first character
	beq.s	ms_exit 		; isn't one
	bsr.s	ioq_pbyt		; there is one, stuff it
	move.b	pt_stuf2(a5),d1 	; is there another?
	beq.s	ms_exit
	bsr.s	ioq_pbyt		; put next byte in queue
	bra.s	ms_exit

ioq_pbyt
	move.l	a3,-(sp)
	move.w	ioq.pbyt,a3
	jsr	(a3)
	move.l	(sp)+,a3
	rts

con_name
	dc.w	11
	dc.b	'CON_2x2a0x0',0

pinf_txt
	dc.w	28
	dc.b	'Pointer Interface not found',$0a

	end
